////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MKB;

import java.util.*;

/**
 *
 * represents an attribute in some relation. <BR> This is
 * meta-knowledge, so the attribute knows only its name, type, and
 * some performance parameters for the QC-Model (size).
  */

public class Attribute {

    /** which "table" do I belong to?. Not sure if we need this. */
    //    private String attTable = null; 
    
    /** the name of this attribute */
    private String attName  = null;

    /** the type of this attribute (as a String) */
    private String attType  = null;

    /**
     * this is the WIDTH of the column, to go with the type. It is not
     * the number of tuples in the attribute. Currently (051999) this
     * is mixed with the physical size of the attribute that we need
     * for QC-computation. We should sort it out later since the
     * logical size could be different from the stored size :( */
    private int attSize=-1;

    /** my parent relation object */
    private Relation parent = null;
    
    
    /**
     * Construct an Attribute Object. Called from parser. With "size"
     * parameter for QC purpose. The "type" is not assigned here since
     * this would have made the MKBParser too complicated (since Type
     * Constraints are defined later than attributes).
     * @author Xin Zhang.  */
    public Attribute(Relation rel, String name, int size){
        attName = name;
        parent = rel;
        attSize = size;
        if (!parent.addAttribute(this))
            System.err.println("MKB.Attribute: Attribute"+rel+"."+name+" could not be added to MKB.");
    }
    
    /** These should eventually be the only remaining
     *   constructors. One assumes that the attribute has the relation
     *   name of its parent, the other allows to overwrite the table
     *   name (for query breakdown) */
    public Attribute (Relation myRel, String myName, 
                      String myType, int mySize) {
        this(myRel, myName, mySize);
        this.setType(myType);
    }
    
    /**
     * returns a string representation of this attribute */
    public String displayAll(){
        return new String(" "+attName+"("+attType+")");
    }
    
    
    /**
     * Dump the Attribute Object.
     * changed for IND finding algorithm July 2001
     * @author Xin Zhang.
     */
    public String toString(){
        return attName;// + ":" + attType + "(" + attSize + ")";
    }
    
    
    /**
     * return a string of the form "attType(attName)". For user
     * interfaces. */
    public String getTypeConstraintAsString() {
        return new String(attType+"("+attName+")");
    }
    
    /**
     * return a TypeConstraint Object. Used for user interface */
    public TypeConstraint getTypeConstraint() {
        return new TypeConstraint(attName, attType);
    }
    
    
    /** get name. */
    public String getName(){
        return attName;
    }
    
    /** get name in the form "ISName.RelName.AttName" */
    public String getFullName(){
        return new String(parent.getFullName()+"."+attName);
    }
    
    /**
     * Return Attribute Name in the form RelName.AttName. Function
     * should be renamed - the name is misleading */
    public String getNoIS(){
        return parent.getName() + "." + attName;
    }

    /**
     * return attribute's type as String
     */
    public String getType(){
	return attType;
    }

    /**
     * set attribute's type as String
     */
    public void setType(String myType){
	attType = myType;
    }

    /**
     * return size (which is just a number passed in the constructor)
     * of this attribute */
    public int getSize() {
        //        if (attSize==-1) 
        //    System.err.println("Attribute.getSize: WARNING. Uninitialized Size. For testing purposes");
        return attSize;
    }
    
    
    /**
     * set size of this attribute */
    public void  setSize(int newSize) {
        if (newSize==-1) 
            System.err.println("Attribute.setSize: WARNING.  Size is -1. For testing purposes");
        attSize = newSize;
    }
    
    /** return the relation object that is this attribute's parent
     */
    public Relation getRelation() {
        return parent;
    }

    /** returns the IS object that is this attribute's parent relation's IS */
    public IS getIS() {
        return parent.getIS();
    }
    
    /**
     * renames attribute in MKB only
     * */

    public boolean changeName(String newName) {
	    attName = newName;
	return true;
    }

}
