package INDModel;

import java.util.Vector;
import DSRG.MKB.Attribute;
import DSRG.MKB.Relation;

/** this class abstracts the concept of Clique over arrays (i.e.,
    simple).  A clique has a number of nodes and a truth value (based
    on whether the corresponding IND is true in the database) */

public class Clique implements Comparable {

    /** the members of this clique. Note that these are not objects
        but ints, and that the nodes are numbered starting with 0 (now
        finally) */
    private int nodes[];

    /** this variable points to the highest index that has a valid
        node inside. Is used so we can use a simple "addVertex" method
        to fill the clique */
     private int ptr;

    /** this marks whether a clique corresponds to a true IND. The bit
        is used in runs of the algo for k>2 */
    private boolean isTrue;


    public Clique(int mySize) {
        nodes = new int[mySize];
        ptr =-1;
        isTrue = false; //proper initialization
        
    }

    public Clique(int vertices[]) {
        this(vertices.length);

        // slow but tested
        for(int i=0; i<vertices.length; i++)
            if (!addCliqueVertex(vertices[i]))
                throw new RuntimeException("Clique error. Node "+vertices[i]+" exists.");
    }

    public int getCliqueSize() {
        return nodes.length;
    }

    public int getCliqueVertex(int i)  {
        return nodes[i];
    }
    
    /** adds a node as long as there is space. Nodes are
        sorted. Duplicate nodes are errors. */
    public boolean addCliqueVertex(int node) {
        if (ptr==nodes.length-1) throw new RuntimeException("Trying to add node to full clique. Clique has only"+nodes.length+" elements.");

        int index=0;
        while(index<=ptr && node>nodes[index]) index++;
        if (ptr!=-1 && node==nodes[index]) return false; // node exists
        for (int i=ptr; i>=index; i--)
            nodes[i+1]=nodes[i];
        nodes[index]=node;
        ptr++;
        return true;
    }

    public void setState(boolean state) {
        isTrue=state;
    }
    
    public boolean getState() {
        return isTrue;
    }

    /** checks if all nodes in subClique exist in Clique. It is
        assumed that the nodes in both cliques are sorted
        lexicographically. */
    public boolean isSupersetOf(Clique subClique) {
        // we can assume that subclique is smaller than clique
        int mySize = this.getCliqueSize();
        if (subClique.getCliqueSize()>mySize) return false;
        int index=0, tmpNode;
        for (int i=0; i<subClique.getCliqueSize();i++) {
            tmpNode=subClique.getCliqueVertex(i);
            while(index<mySize && tmpNode!=nodes[index]) index++;
            if (index==mySize) return false;
        }
        return true;
    }

    public boolean equals(Object otherCliqueObject) {
        Clique otherClique = (Clique)otherCliqueObject;
        if (otherClique.getCliqueSize()!=this.getCliqueSize()) return false;
        for (int i=0; i<this.getCliqueSize(); i++) {
            if (otherClique.getCliqueVertex(i)!=this.getCliqueVertex(i)) return false;
        }
        return true;
    }
    
    /** implementing the Comparable interface. Returns -1 if this
     * object is "less" than the specified object, and so on. "Less"
     * means larger clique, ie. the order is BACKWARDS, so the largest
     * clique comes first! */

     

    public int compareTo(Object other) {
        Clique otherClique = (Clique)other;
        /* // too dangerous
        if (this.getState()==true && otherClique.getState()==false) return -1;
        if (this.getState()==false && otherClique.getState()==true) return 1;
        */

        // the cliques have the same state, so we have to compare their sizes now
        if (otherClique.getCliqueSize()<this.getCliqueSize()) return -1;
        else
            if (otherClique.getCliqueSize()>this.getCliqueSize()) return 1;
            else 
                // equal clique sizes, must compare nodes
                for (int i=0; i<this.getCliqueSize(); i++) {
                    if (otherClique.getCliqueVertex(i)<this.getCliqueVertex(i)) return -1;
                    else 
                        if (otherClique.getCliqueVertex(i)>this.getCliqueVertex(i)) return -1;
                }
        return 0;
    }

    // this function translates a clique to an IND, by using the
    // vector of nodes (unary INDs) in the parameter as a translation
    // table */
    public Ind toIND(Vector nodeToIND) {

        Relation leftR, rightR;
        Vector leftA,rightA; //attribute lists
        Ind tmpUnary;
        leftR=null; rightR=null;
        leftA=new Vector(); rightA=new Vector();
        for (int i = 0; i<this.getCliqueSize();i++) {
            tmpUnary = (Ind)nodeToIND.elementAt(this.getCliqueVertex(i)); 
            if (i==0) {
                leftR=tmpUnary.lRel;
                rightR=tmpUnary.rRel;
            }
            leftA.addElement((Attribute)tmpUnary.lAtts.elementAt(0));
            rightA.addElement((Attribute)tmpUnary.rAtts.elementAt(0));
        }
        // NOTE that we're sorting the INDs alphabetically
        // here again (by attribute names) so it's not the
        // same order as the indices of the unary INDs! Not sure why.
        return new Ind(leftR,leftA,rightR,rightA,Ind.SORTED);
    }

    public String toString() {
        String result = new Integer(nodes.length).toString()+": ";
        for (int i=0; i<=ptr; i++)
            result += new Integer(nodes[i]).toString()+" ";
        result += "("+(isTrue?"true":"false")+")";
        return result+"\n";
    }
}
