package INDModel;

import Discovery;
import DSRG.MKB.*;
import java.util.*;


public class INDCollection {

    private static final boolean DEBUG = false;
    

    /** the original number of attributes in the two relations */

    private int numAtts;

    /** an array of Vectors. The k-th array element represents k-ary
        INDs, so each Vector contains INDs */
    private Vector[] myINDs;

    /** the relations that this INDCollection refers to */
    Relation lRel, rRel;

    /** the instance of the checkIND class that is being used */
    checkInd checkIndInstance;

    /** constructs a new collection of INDs. INDs are stored in an
        array of Vectors, the k-th element in the array therefore is
        the Vector of k-ary INDs.
        <BR>
        This version of the software supports only relations with equal 
        numbers of attributes
        @param lRel_p the left relation to be searched
        @param rRel_p the right relation to be searched
        @param checkIndInstance_p the instance of the checkIND class to be used here
    */

    //============================================================
    //============================================================
    //===================  Constructors    =======================
    //============================================================
    //============================================================

    public INDCollection(Relation lRel_p, Relation rRel_p, checkInd checkIndInstance_p) {
        lRel=lRel_p; rRel=rRel_p;
        if (lRel==null || rRel==null) {
            System.out.println("Could not find one of the Relations. Wrong names?");
            System.exit(0);
        }
        /*
        if (lRel.getAttributes().size()!=rRel.getAttributes().size()) {
            System.out.println("This Version does not support different size relations!");
            System.exit(0);
        }
        */
        // numAtts is the number of attributes of the smaller
        // relation. This is an upper bound for the clique size as
        // well.
        if (lRel.getAttributes().size()<=rRel.getAttributes().size()) 
            numAtts=lRel.getAttributes().size();
        else
            numAtts=rRel.getAttributes().size();


        
        myINDs = new Vector[numAtts+1]; // leaving element 0 free,
        // since there are no 0-ary INDs. Makes calculations easier
        
        for (int i = 1 ; i <=numAtts; i++ ) {
            myINDs[i]= new Vector();
        }

        checkIndInstance=checkIndInstance_p;
    }


    //============================================================
    //============================================================
    //================   Normal Methods    =======================
    //============================================================
    //============================================================

    /** this method checks all INDs of arity k and colors them
        accordingly (myINDs[] is assumed).<BR> Change on 070201: now I
        REMOVE INDs that are not existent. This should reduce the
        search space significantly. <BR> Change on 071101: check only
        INDs that are not already WHITE <BR> Change on 071701: cancel
        first change. leave the black INDs in, we still need
        them. <BR>Change on 090901: restrict number of true
        INDs. First get all INDs, together with their "support"
        (currently defined as the number of distinct values in the
        smaller attribute) and the keep only the x INDs with the
        highest support (x being the global variable MAX_NUM_NODES for
        unary INDs and MAX_NUM_EDGES for binary INDs).

        @param theINDs the vector  of INDs to discover
        @return the number of INDs whose state has been changed to WHITE (i.e., the true ones) */
    public int checkKAryInds(Vector theINDs) {
        double support;
        boolean isTrue;
        Ind tmpInd;
        int counter=0;
        
        for (Iterator iter=theINDs.iterator();iter.hasNext();) {
            tmpInd = (Ind)iter.next();
            if (tmpInd.getState()==Ind.WHITE) continue;
            support = checkIndInstance.Main(tmpInd);
            isTrue = (support>0?true:false);
            if (isTrue) {
                tmpInd.setState(Ind.WHITE);
                tmpInd.setSupport(support);
                counter++;
            }
            else tmpInd.setState(Ind.BLACK);
        }

        // assuming all INDs have the same arity.
        int k = ((Ind)theINDs.elementAt(0)).getArity();

        // all this is done only for unary and binary INDs hack:
        // 092501, now done for all INDs, so the hypergraphs don't get
        // too big
        if (true/*k<=2*/) {
            
            // now need to sort all INDs by their
            // support. Insertionsort should be fast enough I would
            // say. Very primitive, too. Descending order.
            int j;
            for (int i=1; i<theINDs.size(); i++) {
                tmpInd = (Ind)theINDs.elementAt(i);
                j=i-1;
                while (j>=0 && 
                       tmpInd.getSupport() 
                       > 
                       ((Ind)theINDs.elementAt(j)).getSupport()
                       ) {
                    theINDs.setElementAt(theINDs.elementAt(j),j+1);
                        j=j-1;
                }
                theINDs.setElementAt(tmpInd, j+1);
            }
            
            // now count the first x valid INDs and set the rest to BLACK
            int countDown= (k==1)?Discovery.MAX_NUM_NODES:Discovery.MAX_NUM_EDGES;
            int index,rejectcounter=0;
            for (index =0; index < theINDs.size() && countDown>0; index++) {
                if ( ((Ind)theINDs.elementAt(index)).getState()==Ind.WHITE)
                    countDown--;
            }
            if (index < theINDs.size()) {
                counter = (k==1)?Discovery.MAX_NUM_NODES:Discovery.MAX_NUM_EDGES;
                for (;index<theINDs.size(); index++) {
                    tmpInd = (Ind)theINDs.elementAt(index);
                    if (tmpInd.getState()==Ind.WHITE) rejectcounter++;
                    tmpInd.setState(Ind.BLACK);
                }
            }
            System.out.println("\n"+rejectcounter+" INDs rejected due to MAX_NUM bounds");
        

        }
        
        /*
        //added 070201, cancelled 071701
        for(int i=0; i<theINDs.size(); i++) {
            if(((Ind)theINDs.elementAt(i)).getState()!=Ind.BLACK) {
                result.addElement(theINDs.elementAt(i));
            }
        }
        */

        return counter; // careful with the objects here
    }
    
    /** removes invalid INDs from a vector, for performance
        reasons. The returned vector is a shallow clone (same objects,
        different pointers) */
    public Vector removeInvalidINDs(Vector theINDs) {
        Vector result = new Vector();
        for(int i=0; i<theINDs.size(); i++) {
            if(((Ind)theINDs.elementAt(i)).getState()==Ind.WHITE) {
                result.addElement(theINDs.elementAt(i));
            }
        }
        return result;
    }

    public Vector generateUnaryINDs() {
        Vector result = new Vector();
        
        Vector lAtts = lRel.getAttributes();
        Vector rAtts = rRel.getAttributes();
        
        // each attribute from rel1 can be included in each attribute in rel2
        Ind tmpInd; Vector tmpVec1, tmpVec2;
            for (int i = 0; i<lAtts.size(); i++)
                for (int j = 0; j < rAtts.size(); j++) {
                    tmpVec1 = new Vector();
                    tmpVec1.addElement(lAtts.elementAt(i));
                    tmpVec2 = new Vector();
                    tmpVec2.addElement(rAtts.elementAt(j));
                    tmpInd = new Ind(lRel,
                                     tmpVec1,
                                     rRel,
                                     tmpVec2
                                     );
                    result.addElement(tmpInd); 
                }
        return result;
    }
    public Vector generateBinaryINDs(Vector unaryINDs){
        Vector result = new Vector();

            // find 2-attribute inclusion dependencies
            

            // 071401 this is now too complicated. but we will keep it
            // for the time being.
            Ind oldInd, newInd, tmpInd;
            for (int i = 0 ; i < unaryINDs.size(); i++)
                for (int j = i+1 ; j < unaryINDs.size(); j++) {
                    //if (i==j) continue;
                    oldInd = (Ind)unaryINDs.elementAt(i);
                    newInd = (Ind)unaryINDs.elementAt(j);
                    // need to avoid adding new atts that already exist
                    if (oldInd.lAtts.indexOf(newInd.lAtts.elementAt(0))!=-1
                        ||
                        oldInd.rAtts.indexOf(newInd.rAtts.elementAt(0))!=-1
                        ) 
                        continue;
                    // not adding INDs that are based on INDs colored BLACK
                    if (oldInd.getState()==Ind.BLACK
                        ||
                        newInd.getState()==Ind.BLACK
                        ) 
                        continue;
                    //else
                    tmpInd = oldInd.addIndAndSort(newInd);
                    result.addElement(tmpInd);
                }

        // ok (072401), I trust that this is not going to happen any
        // longer. It slows down the algorithm too much
        /*

        //  remove duplicates, the primitive way
        Vector indSet = new Vector();

        // this shouldn't happen anymore but will remain in here for sanity checks

        for (Iterator iter = result.iterator();
             iter.hasNext();) {
            tmpInd = (Ind)iter.next();
            if (!indSet.contains(tmpInd)) {
                indSet.addElement(tmpInd);
                //                result++; // counting INDs added
            }
            else throw new RuntimeException("Hmm, we have generated a duplicate IND: "+tmpInd);
        }
        result = indSet;
        */

        return result;
    }
   

    /** checks all two-ary INDs for cliques and then test the cliques
        as large-ary INDs. Return all cliques, but mark them according
        to whether they're true in the database
        @param k the arity of INDs we want to use as base
        objects. This corresponds to the k in the k-Hypergraph.
        @return vector of cliques found (Clique objects).  */

    public TreeSet generateCliquesAndCheck(Vector nodes, Vector edges, int k) {
        int numNodes=nodes.size();
        Ind tmpInd;
        TreeSet cliques, packedCliques;
        if (k==2) {
            // we assume that k is 2 here. I.e., the algo will at the
            // moment work only for binary INDs as base objects. Each
            // binary IND is an edge on a graph whose nodes are the unary
            // INDs.  After running the clique algorithm, we have a vector
            // whose elements are again vectors. Each of those vectors
            // contains the indices of those unary INDs whose binary pairs
            // form cliques. I.e., each such set can be tested as a
            // high-arity IND. 
        
            // as in the original algorithm, everything here is one-based!!
            boolean theGraph[][] = new boolean[numNodes][numNodes];
            for (int i=0; i<numNodes; i++)
                for (int j=0; j<numNodes; j++) 
                    if (i==j)
                        theGraph[i][j]=true;
                    else
                        theGraph[i][j]=false;
            int index1, index2;
            for(Iterator iter = edges.iterator(); iter.hasNext();) {
                tmpInd = (Ind)iter.next();
                index1 = nodes.indexOf(tmpInd.removeIndAt(1));
                index2 = nodes.indexOf(tmpInd.removeIndAt(0));
                // weird method to get the sub-INDs, but it's for the trash anyway
                theGraph[index1][index2]=true;
                theGraph[index2][index1]=true; 
            }
            //we should now have the graph. Let's check that.
            if (DEBUG)
                for (int i=0; i< numNodes; i++) {
                    for (int j=0; j< numNodes; j++)
                        System.out.print(theGraph[i][j]?1:0);
                    System.out.println();
                }
        
            cliques = findCliques.findCliques(theGraph);
        
        } // if(k==2)

        else { // k>2, so we need to do all of the above in Hypergraphs
            // first, construct the Hypergraph.
            if (DEBUG) System.out.println("Constructing "+k+"-hypergraph.");
            int indices[] = new int[k];
            KHypergraph theGraph = new KHypergraph(k, numNodes);
            // go through all k-ary INDs. Each of them is an edge in the hypergraph
            for(Iterator iter = edges.iterator(); iter.hasNext();) {
                tmpInd = (Ind)iter.next();
                if (DEBUG) System.out.print("adding edge: ");
                for (int i =0; i<k; i++) {
                    indices[i]=nodes.indexOf(tmpInd.unarySubInd(i));
                    if (DEBUG) System.out.print(indices[i]+" ");
                }
                if (DEBUG) System.out.println();
                theGraph.addEdge(indices);
            }

            // now, find the cliques. 
            cliques = findCliques.findHyperCliques(theGraph);
            
        }

        System.out.println("Generated "+cliques.size()+ " cliques.");

            // now we have the cliques, and we need to check them now.
            // for that, we iterate through the cliques and first need to
            // translate them back into INDs. 
        
            Relation leftR, rightR;
            Vector leftA,rightA; //attribute lists
            Clique tmpClique;
            Ind tmpUnary,newInd;
            int newArity;

            // now testing up to MAX_CLIQUES_TESTED cliques, in order
            // of size (number of nodes in the clique). Retaining up
            // to MAX_CLIQUES_KEPT valid ones, and if not so many are
            // valid, also keeping invalid ones up to a total number
            // of MAX_CLIQUES_KEPT
            int counter=Discovery.MAX_CLIQUES_TESTED;
            packedCliques = new TreeSet();

            // first, collect up to MAX_CLIQUES_KEPT true cliques. Stop after MAX_CLIQUES_TESTED (i.e., if counter is zero)
            for(Iterator iter=cliques.iterator(); iter.hasNext();) {
                tmpClique=(Clique)iter.next();
                if (DEBUG) System.out.println("Current Clique: "+tmpClique);
                newInd = tmpClique.toIND(nodes);

                if (DEBUG) System.out.println("Clique to IND gives: "+newInd);
                newArity = newInd.getArity();
                double support;
                if (newArity>k) { //cuz we already have all the smaller INDs
                    support = checkIndInstance.Main(newInd);
                    if (support>0) {
                        tmpClique.setState(true);
                        newInd.setState(Ind.WHITE);
                        newInd.setSupport(support);
                        myINDs[newArity].addElement(newInd);
                        if (Discovery.DEBUG) System.out.println("added to myINDs["+newArity+"]");
                    }
                    else {
                        /* we're not removing cliques anymore since we'll also need
                           the falsified ones
                           cliques.removeElementAt(cliqueIndex);
                           cliqueIndex--;
                        */
                    }
                }
                else {// 
                    tmpClique.setState(true); // trivially true
                }
                // otherwise, the clique is false.

                if (tmpClique.getState()==true) packedCliques.add(tmpClique);
                if (packedCliques.size()>=Discovery.MAX_CLIQUES_KEPT) break;
                if (counter--==0) break;
            }

            // now add cliques that tested negatively, again up to
            // MAX_CLIQUES_KEPT or MAX_CLIQUES_TESTED. We don't need
            // to check the DB again since we already have all the
            // necessary information (i.e., either we already have
            // MAX_CLIQUES_KEPT cliques or we have at least tested
            // MAX_CLIQUES_TESTED cliques so the first
            // MAX_CLIQUES_TESTED cliques have a correct state. 
            
            if (packedCliques.size()<Discovery.MAX_CLIQUES_KEPT) {
                // we have a little more space for invalid cliques.
                // so we take the first few invalid cliques and add them here
                for (Iterator iter=cliques.iterator(); iter.hasNext();) {
                    tmpClique=(Clique)iter.next();
                    if (tmpClique.getState()==false)
                        packedCliques.add(tmpClique);
                    if (packedCliques.size()>=Discovery.MAX_CLIQUES_KEPT) break;
                    if (counter--==0) break;
                }
            }
    
            // now we have no more than MAX_CLIQUES_KEPT cliques,
            // mostly valid ones (if there were any).
        return packedCliques;
    }
    
    /** this forms all k-ary subsets of all falsified cliques in
        "cliques" and puts them into myINDs[k]. We assume that the
        only INDs of arity k already in myINDs are all tested (white)
        from a previous step. So we will test for containment ("is
        that IND already there and white?") but not overwrite
        anything. We will see if there is a need for more states. 
        @param level the arity of the subsets to be generated from the
        cliques (only the false ones, of course!)
        @param cliques the Vector of cliques from which we check */

    public Vector generateIndsFromCliques(int level,Vector nodes, TreeSet cliques) {
        if (DEBUG) System.out.println("Generating "+level+"-level INDs from "+cliques);
        // test all cliques
        Clique tmpClique;
        int cSize;
        Ind newInd,tmpUnary;
        Vector leftA,rightA;
        Relation leftR,rightR;
        Vector result=new Vector();
        if (cliques.size()>Discovery.MAX_CLIQUES_KEPT)
            throw new RuntimeException("Hmm. Too many cliques here. Impossible. "+cliques);
        for (Iterator iter = cliques.iterator(); iter.hasNext();) {
            tmpClique = (Clique)iter.next();
            cSize = tmpClique.getCliqueSize();
            if (cSize<level) continue; // not this one

            if(tmpClique.getState()==true) continue; 
            // not this one either, since all it's sub-INDs must be true.
            if (binomial(cSize,level)>Discovery.MAXINDS_GENERATED) {
                System.out.println("Can not generate "+level+"-ary INDs from clique: "+tmpClique.toString()+" (max. "+Discovery.MAXINDS_GENERATED+")");
            }
            else {
                // we must generate all level-ary subsets from a cSize-ary
                // clique. 
                if (DEBUG) System.out.println("generating clique-subsets for "+tmpClique);
                
                boolean alreadyThere;
                
                // we use the same subset-generating algorithm
                // again. cSize plays the role of "nn", the "degree"
                // is the target clique size which is "level" here.
                
                //let's rename the variables. I should make this an
                //Iterator but have to first figure out how to do
                //that.
                
                int indices[] = new int[level]; 
                // for the indices of the subset elements
                int translatedIndices[] = new int[level];
                // since the indices are n from 0..level, and not the actual nodes.

                // initialize indices
                for (int i = 0; i<level; i++)
                    indices[i]=i;
                
                int i_ptr = level-1;
                // now form subsets 
                
                do {
                    
                    /*
                      if (DEBUG) { 
                        System.out.print("Subset: ");
                        for (int m=0; m<level; m++)
                            System.out.print(indices[m]+" ");
                        System.out.println();
                    }
                    */
                    
                    // some subsets generated here could already be
                    // subsumed by other true cliques. Currently, we
                    // need to filter them out. It would be better if
                    // we could avoid their generation. 
                    
                    // we (unfortunately) need to cycle through all
                    // true cliques and check if our current subset is
                    // included in any of them

                    for(int i=0; i<level; i++)
                        translatedIndices[i]=tmpClique.getCliqueVertex(indices[i]);

                    boolean alreadySubsumed=false;
                    Clique newClique=new Clique(translatedIndices);
                    Clique localClique; int localCSize;

                    for (Iterator iter2 = cliques.iterator(); iter2.hasNext();) {
                        localClique = (Clique)iter2.next();
                        localCSize = localClique.getCliqueSize();
                        if (localCSize<=level) continue; // not this one. Could be improved by sorting cliques
                        if(localClique.getState()==false) continue;  // here, we need only the true ones
                        //if (DEBUG) System.out.println("Checking if "+localClique+" is superset of "+newClique);
                        if(localClique.isSupersetOf(newClique)) {
                            alreadySubsumed=true;
                            break;
                        }
                    }

                    // hmm, even if the IND is subsumed, we need it
                    // for the clique-finding. So we have to leave it
                    // in, but don't need to test it.
                        leftR=null;rightR=null;leftA=new Vector(); rightA=new Vector();
                    

                        // we have a subset. now turn it into an IND
                        for (int dim=0; dim<level; dim++) {
                            tmpUnary = (Ind)nodes.elementAt(translatedIndices[dim]); 
                            if (dim==0) {
                                leftR=tmpUnary.lRel;
                                rightR=tmpUnary.rRel;
                            } 
                            leftA.addElement((Attribute)tmpUnary.lAtts.elementAt(0));
                            rightA.addElement((Attribute)tmpUnary.rAtts.elementAt(0));
                        }
                        
                        newInd = new Ind(leftR,leftA,rightR,rightA);
                        //if (DEBUG) System.out.println("New sub-IND: "+newInd);

                        alreadyThere=false;
                        for(int m=0; m<result.size();m++) {
                            if(((Ind)result.elementAt(m)).equals(newInd) ){
                                alreadyThere=true;
                                break;
                            }
                        }
                        if (!alreadyThere) {
                            if (alreadySubsumed) {
                                newInd.setState(Ind.WHITE);
                                if (DEBUG) System.out.println("Clique is already subsumed, so IND must be true.");
                            }
                            result.addElement(newInd);
                            if (DEBUG) System.out.println("New IND generated");
                        }
                        else if (DEBUG) System.out.println("IND/Clique existed already. Not added.");
                
                    // generate next subset
                    i_ptr=level-1;
                    indices[i_ptr]++;
                    
                    // if out of bounds, push up the carry-overs
                    for (i_ptr=level-1;i_ptr>0;i_ptr--) {
                        if (indices[i_ptr]==cSize-(level-1-i_ptr)) {// index at position k cannot be larger than k+(cSize-level)
                            if (i_ptr==0) break;
                            else {
                                indices[i_ptr-1]++;
                                indices[i_ptr]=-1;
                            }
                        }
                    }
                    for(i_ptr=1; i_ptr<level; i_ptr++) {
                        if (indices[i_ptr]==-1)
                            indices[i_ptr]=indices[i_ptr-1]+1;
                    }
                }
                while (indices[0]<cSize-(level-1));
                
            }
        }
        return result;
    }


    //=======================================================
    //=======================================================
    //============  get/set-methods =========================
    //=======================================================
    //=======================================================


    /** returns the IND-collection: an array of vectors, containing INDs */
    public Vector[] getINDCollection() {
        return myINDs;
    }

    /** returns the number of attributes of the two relations that
     these INDs are from.  */

    public int getNumAtts() {
        return numAtts;
    }






    //============================================================
    //============================================================
    //==============   static computation functions   ============
    //============================================================
    //============================================================

    /** computes total number of all possible INDs for a certain
        number of attributes
        @param k number of attributes in relation 
    */
    static public int totalMaxNumberOfINDs(int k) {
        int result =0;
        for (int i = 1; i <=k ;i++)
            result += maxNumberOfINDs(k,i);
        return result;
    }

    /** computes number of possible INDs for k and i 
        @param k the number of attributes in each relation
        @param i the arity of the INDs to count */
    static public int maxNumberOfINDs(int k, int i) {
        if (k<i) return -1;
        int tmp=binomial(k,i);
        return tmp*tmp*fact(i);
    }

    /** binomial coefficient of n and k */
    static private int binomial(int n, int k) {
        if (n<k) return -1; // error

        int result =1;
        for(int i = 0; i< k; i++)
            result *= n-i;
        result /= fact(k);
        return result;
    }

    /** factorial */
    static private int fact(int n) {
        int result=1;

        for(int i =2; i<=n; i++) 
            result *= i;
        return result;
    }

    //============================================================
    //============================================================
    //========================  debugging help    ================
    //============================================================
    //============================================================

    public String toString() {
        String result = new String();
        for (int i = 1 ; i <=numAtts; i++) {
            result += new Integer(i).toString()+":\n";
            for (Iterator iter=myINDs[i].iterator();iter.hasNext();)
                result += ((Ind)iter.next()).toString()+"\n";
            result +="\n";
        }
        return result;
    }
}

