package INDModel;

import java.util.*;
import DSRG.MKB.*;


/** An inclusion dependency (IND). An IND contains two references two
relations, and two Vectors containing references to Attributes. Also, each IND can be in one of three states: GREY (undiscovered), BLACK (false), and WHITE (true). 
*/

public class Ind {

    /** a constant for sorted concatenation of INDs */
    public static final boolean SORTED = true;
    /** Ind is undiscovered */
    public static final int GREY = 1;
    /** Ind is discovered and false */
    public static final int BLACK = 2;
    /** Ind is discovered and true */
    public static final int WHITE = 3;

    Relation lRel;
    Relation rRel;
    Vector lAtts=new Vector();
    Vector rAtts=new Vector();

    /** the status of the Ind. It is initialized to GREY. */
    int state=GREY;

    /** the "support" of the Ind. Simply the number of distinct values
        in the smaller attribute */

    private double support;

    //========================================================
    //========================================================
    //===============  Constructors    =======================
    //========================================================
    //========================================================


    /** copy constructor copies REFERENCES to relations, but CLONES
        the attribute vectors (such that the new vector still points
        to the same attributes, but is a different Vector) */

    public Ind(Ind tmpInd) {
        lRel = tmpInd.lRel;
        rRel = tmpInd.rRel;
        lAtts = (Vector)tmpInd.lAtts.clone();
        rAtts = (Vector)tmpInd.rAtts.clone();
    }

    /** creates a new IND, now always sorted (same as the other constructor)
        @param l left Relation
        @param a1 Vector of Attribute, left projection list
        @param r right Relation
        @param a2 Vector of Attribute, right projection list */
    public Ind(Relation R1, Vector a1, Relation R2, Vector a2) {
        this(R1,a1,R2,a2,Ind.SORTED);
    }
    
    /** this adds a sorted inclusion dependency. The first attribute
        vector is being sorted, and the second string follows the
        permutation of the first one. <BR><I> Currently, this
        constructor does NOT make copies of the attribute vectors but
        rather sorts them and them assigns their references to class
        member variables</i>
        @param R1 left Relation
        @param a1 Vector of Attribute, left projection list
        @param R2 right Relation
        @param a2 Vector of Attribute, right projection list
        @param sortme if this is equal to  Ind.SORTED, the IND is added with an alphabetically sorted attribute list */
    public Ind(Relation R1, Vector a1, Relation R2, Vector a2, boolean sortme) {
        lRel = R1;
        lAtts = a1;
        rRel = R2;
        rAtts = a2;
        if (sortme==SORTED) {
            int j;
            Attribute ltmp,rtmp;
            // insertion sort. hopefully not too slow for 400 attributes...
            for (int i = 1 ; i<a1.size(); i++) {
                j = i-1;
                ltmp = (Attribute)a1.elementAt(i);
                rtmp = (Attribute)a2.elementAt(i);
                while (j>=0 && ltmp.toString().compareTo(a1.elementAt(j).toString())<0) {
                    a1.setElementAt(a1.elementAt(j),j+1);
                    a2.setElementAt(a2.elementAt(j),j+1);
                    j--;
                }
                a1.setElementAt(ltmp,j+1);
                a2.setElementAt(rtmp,j+1);
            }
            lAtts = a1;
            rAtts = a2;
        }
    }


    //==================================================
    //==================================================
    //========        Normal Methods   =================
    //==================================================
    //==================================================

    
    /** adds an Ind to the right of this one and returns the result as
        a new object */
    public Ind xaddInd(Ind newInd) {
        Ind tmpInd = new Ind(this);
        for (Iterator iter = newInd.lAtts.iterator(); iter.hasNext();)
            tmpInd.lAtts.addElement(iter.next());
        for (Iterator iter = newInd.rAtts.iterator(); iter.hasNext();)
            tmpInd.rAtts.addElement(iter.next());
        return tmpInd;
    }

    /** removes a sub-Ind from this one (set-minus). If the Ind (or
        one of its components) does not exist in this one, null is
        returned. Otherwise, function returns the new Ind. */
    public Ind xremoveInd(Ind removeInd) {
        if (!this.contains(removeInd))
            return null; // Ind is not contained
        Ind tmpInd = new Ind(this);
        Attribute tmpLAtt, tmpRAtt;
        int index=0;
        //else
        for (int i =0; i< removeInd.getArity(); i++) {
            index = 0;
            tmpLAtt = (Attribute)removeInd.lAtts.elementAt(i);
            tmpRAtt = (Attribute)removeInd.rAtts.elementAt(i);
            do {
                index = this.lAtts.indexOf(tmpLAtt,index);
            }
            while (index!=-1 &&
                   this.rAtts.elementAt(index)!=tmpRAtt
                   // at this point we know that the pair exists, so
                   // this must match once
                   );
                
            tmpInd.lAtts.removeElementAt(index);
            tmpInd.rAtts.removeElementAt(index);
            
        }
        return tmpInd;
    }

    /** removes a unary sub-Ind at a certain position (which must be
        smaller than this Ind's arity */
    public Ind removeIndAt(int k) {
        if (k>=this.getArity())
            return null;
        //else
        Ind tmpInd = new Ind(this);
        tmpInd.lAtts.removeElementAt(k);
        tmpInd.rAtts.removeElementAt(k);
        return tmpInd;
    }
    
    /** returns a (newly generated) unary sub-IND at index k */
    public Ind unarySubInd(int k) {
        if (k>=this.getArity())
            return null;
        //else
        Vector rA=new Vector(), lA=new Vector();
        rA.addElement(rAtts.elementAt(k));
        lA.addElement(lAtts.elementAt(k));
        return new Ind(lRel,lA,rRel,rA);
    }


    /** does the same as addIndAndSort, except that only those
    attribute pairs frmo the new IND are added that do not already
    exist in this IND */
    public Ind merge(Ind other) {
        Ind tmpInd = new Ind(this);
        Vector tmpLeft  = tmpInd.lAtts;
        Vector tmpRight = tmpInd.rAtts;
        Attribute left, right;
        // we assume that both sides of the IND have an equal number
        // of attributes
        Iterator iter2 = other.rAtts.iterator();
        for (Iterator iter = other.lAtts.iterator(); iter.hasNext();) {
            left = (Attribute)iter.next();
            right = (Attribute)iter2.next();
            if(!tmpLeft.contains(left) ||
               (tmpLeft.contains(left) 
                && !((Attribute)tmpRight.elementAt(tmpLeft.indexOf(left))).equals(right)
                )
               ) {
                tmpLeft.addElement(left);
                tmpRight.addElement(right);
            }
        }
        
        tmpInd = new Ind(other.lRel,tmpLeft,
                             other.rRel,tmpRight,
                             true); // concatenate and sort
        return tmpInd;
    }


    /** adds an Ind to the right of this one and returns the result as
        a new object but sorts the IND afterwards (alphabetically by
        attribute names). This ensures that two equal INDs are in fact
        recognized as equal */
    public Ind addIndAndSort(Ind newInd) {
        Ind tmpInd = new Ind(this);
        Vector tmpLeft  = tmpInd.lAtts;
        Vector tmpRight = tmpInd.rAtts;
        for (Iterator iter = newInd.lAtts.iterator(); iter.hasNext();)
            tmpLeft.addElement(iter.next());
        for (Iterator iter = newInd.rAtts.iterator(); iter.hasNext();)
            tmpRight.addElement(iter.next());
        

        tmpInd = new Ind(newInd.lRel,tmpLeft,
                             newInd.rRel,tmpRight,
                             true); // concatenate and sort
        return tmpInd;
    }

    //=======================================================
    //=======================================================
    //============  get/set-methods =========================
    //=======================================================
    //=======================================================

    public void  setSupport(double mysupport) {
        support = mysupport;
    }

    public double getSupport() {
        return support;
    }

    public int getState() {
        return state;
    }

    public void setState(int newState) {
        state = newState;
    }

    /** the arity of an IND is the number of attributes on either
        side. They must be equal */
    public int getArity() {
        if (this.lAtts.size()!=this.rAtts.size()) {
            System.err.println("arity mismatch in IND: "+this.toString());
            System.exit(0);
        }
        return this.lAtts.size();
    }
    
    //=======================================================
    //=======================================================
    //===========    Java-typical overwritten methods = =====
    //=======================================================
    //=======================================================

    public String toString() {
        String status;
        switch (state) {
            case BLACK : status="black"; break;
            case WHITE : status="white"; break;
        default : status = "grey";
        }
            
        return lRel.getName() + " ("+lAtts+") <== " + rRel.getName() +" ("+rAtts+")"+" Status: "+status+" Support: "+support+"\n";
    }

    public boolean equals(Object tmp) {
        return this.equals((Ind)tmp);
    }

    public boolean equals(Ind tmp) {

        if(getArity()!=tmp.getArity()) return false;
        if (!lRel.equals(tmp.lRel)) return false;
        if (!rRel.equals(tmp.rRel)) return false;
        
        // if lAtts and rAtts are the same objects, the INDs must be equal
        if(lAtts==tmp.lAtts && rAtts==tmp.rAtts) return true;

        // else, we need to really compare both INDs
        // we can rely on the INDs being sorted.
        
        if (lAtts.size()!=rAtts.size()) return false;
        for(int i=0; i<lAtts.size();i++)
            if (lAtts.elementAt(i)!=tmp.lAtts.elementAt(i) 
                ||
                rAtts.elementAt(i)!=tmp.rAtts.elementAt(i)
                )
                return false;
        return true;
    }
    
    /** this function determines if a passed IND is contained in this
    object. This is accomplished by comparing each pair of attributes
    in tmp, and trying to find the same pair in "this". <B>It is
    assumed that tmp.lAtts.size()==tmp.rAtts.size()</B> */
    public boolean contains(Ind tmp) {
        Attribute tmpLAtt,tmpRAtt;
        int whereIsTmpLAtt;
        for (int i = 0; i<tmp.getArity(); i++) {
            tmpLAtt = (Attribute)tmp.lAtts.elementAt(i);
            tmpRAtt = (Attribute)tmp.rAtts.elementAt(i);
            whereIsTmpLAtt = this.lAtts.indexOf(tmpLAtt);
            if ( whereIsTmpLAtt==-1 || 
                 this.rAtts.elementAt(whereIsTmpLAtt)!=tmpRAtt
                 ) {
                return false;
            }
        }
        // only if all attributes in tmp are matching somewhere in
        // this, the Ind tmp is contained
        return true;
    }

}
