package INDModel;

import java.util.*;

/** This class declares a Hypergraph in which all edges have exactly k
    nodes. We call it k-Hypergraph. The arity (degree) must be passed
    in the constructor.

<P> Since edges are sets of nodes, permuting the node indices in an
edge does not change the graph. Therefore, edges are always stored as
SORTED arrays of integers.  */

public class KHypergraph {

    private final static boolean DEBUG=false;

    private int arity;

    private int nodes;
    /** the vector of edges contains a vector of KEdges of arity
        "arity". Each such array contains the indices of all nodes in
        this edge of the k-hypergraph in ascending order. Furthermore,
        edges are stored in lexicographic order themselves. This order
        is maintained since edges can only be added through addEdge() */
    Vector edges = new Vector();

    /** the constructor needs the arity (i.e., the number of nodes in
        each edge) for this k-Hypergraph and the (initial) number of
        nodes in this k-Hypergraph. 
    @param arityP the arity of this hypergraph. Cannot change.
    @param nodesP the number of nodes. This parameter is currently
    reported back by getNumNodes() but is otherwise not maintained
    inside the class. */
    public KHypergraph(int arityP, int nodesP) {
        arity = arityP;
        nodes=nodesP;
    }
    
    
    /** returns arity of this k-Hypergraph. Naturally, the arity can
        not be changed. */
    public int getArity() {
        return arity;
    }

    /** returns the initial number of nodes that was passed in the
        constructor. This may have nothing to do with the actual
        number of nodes in the graph. We may have to change that at
        some point. */
    public int getNumNodes() {
	return nodes;
    }

    /** returns the degree of a node. The degree is defined as the
        number of edges that contain this node. The node is given as a
        0-based integer. */
    public int getDegree(int index) {
	int count =0;
	for (Iterator iter=edges.iterator(); iter.hasNext();)
	    if(((KEdge)iter.next()).contains(index)) count++;
	return count;
    }
    
    /** returns an int-array of all node degrees. The array has as
        many elements as the graph has nodes. <B> we rely on the
        correctness of the parameter <code>nodes</code>, passed in the
        constructor.</B> */
    public int[] getAllDegrees() {
        int[] result = new int[nodes];
        KEdge tmpEdge=null;
	for (Iterator iter=edges.iterator(); iter.hasNext();) {
            tmpEdge = (KEdge)iter.next();
            for (int i =0; i<arity; i++)
                result[tmpEdge.nodeAt(i)]++;
        }
        return result;
    }

    /** returns the degrees of all nodes in the subgraph induced by
        the set of nodes passed as parameter. <B>CHANGE: we now ignore
        -1 entries in subnodes as if they weren't there :)</B>
        @return an array of degrees, with the node-numbers as indices
        (i.e., the array has subnodes.length elements) */
    public int[] getAllDegreesOfSubgraph(int[] subnodes) {
	KEdge subEdge = new KEdge(subnodes);
	int [] inverseIndex=new int[nodes];
	for(int i=0; i<nodes; i++)
	    inverseIndex[i]=-1;
	for(int i=0; i<subnodes.length; i++)
	    if (subnodes[i]>=0)
		inverseIndex[subnodes[i]]=i;
        
        int[] result = new int[subnodes.length];
        KEdge tmpEdge=null;
        int[] tmpArray;
	for(Iterator iter=edges.iterator(); iter.hasNext();) {
            tmpEdge = (KEdge)iter.next();
            if(tmpEdge.isSubsetOf(subEdge))
		for(int i =0; i<arity; i++)
                    result[inverseIndex[tmpEdge.nodeAt(i)]]++;
        }
	return result;
    }

    public Vector getAllEdgesOfSubgraph(int[] subnodes) {
        Vector result = new Vector();
        KEdge bigEdge = new KEdge(subnodes);
        KEdge tmpEdge;
        for (Iterator iter=edges.iterator(); iter.hasNext();) {
            tmpEdge = (KEdge)iter.next();
            if (tmpEdge.isSubsetOf(bigEdge))
                result.addElement(tmpEdge);
        }
        return result;
    }

    /** adds an edge that is given by an array of its node indices.
@return false on the following error conditions: indices[] does not
have "arity" elements, edge already exists; true otherwise. */
    public boolean addEdge(int indices[]) {
        if (indices.length!=arity) throw new RuntimeException("Wrong arity. Trying to add edge of arity "+new Integer(indices.length).toString()+" to "+new Integer(arity).toString()+"-Hypergraph");
        KEdge tmpEdge = new KEdge(indices);
        if (DEBUG) System.out.println("Adding Edge: "+tmpEdge);
	int numEdges=edges.size();
        if (numEdges==0) {
            edges.addElement(tmpEdge);
            return true;
        }
        // linear search for insertion point... binary search is too painful

        boolean inserted=false;
        int oldSize=edges.size();
        for(int i =0; i< oldSize; i++) {
            if (tmpEdge.lessThanOrEqual((KEdge)edges.elementAt(i))) {
                if (tmpEdge.equals(((KEdge)edges.elementAt(i))))
                    throw new RuntimeException("Edge "+tmpEdge.toString()+" exists.");
                else  { // tmpEdge is now greater than element at index
                    edges.insertElementAt(tmpEdge,i);
                    inserted=true;
                    break;
                }
            }
        }
        if (!inserted) edges.addElement(tmpEdge);
        
        
        /* // even this was too hard. now doing it very slowly
        int dim = -1; int index=0; int thisNode=-1,prevNode=-1;
        do {
            dim++;
	    if(dim==arity) break; // i.e., dim too big for edge
            thisNode=tmpEdge.nodeAt(dim);
            while(index<numEdges 
                  && ((KEdge)edges.elementAt(index)).nodeAt(dim)<thisNode
                  ) index++;
            prevNode=thisNode;
        }
        while(index<numEdges 
              && ((KEdge)edges.elementAt(index)).nodeAt(dim)==thisNode
              && (dim==0 || ((KEdge)edges.elementAt(index)).nodeAt(dim-1)==prevNode)
              );

        if (dim==arity)
	    throw new RuntimeException("Edge "+tmpEdge.toString()+" exists.");
	
        
        edges.insertElementAt(tmpEdge,index);
        */
        return true;
    }

    /** this function deletes the node at index 1 and removes all
        edges that contain this node. If the index of the node to be
        removed is out of range, false is returned, otherwise
        true. <B> NOT IMPLEMENTED YET </B> */
    public boolean deleteNode(int index) {
        throw new RuntimeException("function not implemented");
    }
    
    public boolean isClique(int subset[]) {
        int target = binomial(subset.length-1,arity-1);
        int degrees[] = getAllDegreesOfSubgraph(subset);
        for(int i=0; i<subset.length; i++) {
            if (degrees[i]!=target) return false;
        }
        return true;
    }
    
    /** tests if a given set of nodes forms a clique. In order to do
        this, the graph is reduced to only the nodes in the (sorted!!)
        array indices, and it is checked if each node degree is still
        equal to the length of the subset[] parameter.
        @param subset the subset of nodes that is to be tested for
        cliqueness. MUST BE SORTED! */
    public boolean xisClique(int subset[]) {
	int tmpNode, j;
	// sort clique first. is needed for new algorithm
        for (int i = 1 ; i<subset.length; i++) {
            j = i-1;
            tmpNode = subset[i];
            while (j>=0 && tmpNode<subset[j]) {
                subset[j+1]=subset[j];
                j--;
            }
            subset[j+1]=tmpNode;
        }

        if (DEBUG) System.out.println("Trying to see if "+new KEdge(subset).toString()+" is a clique");
        int nn = subset.length; // target clique size

        if (nn<arity) throw new RuntimeException("Trying to check for "+arity+"-clique in set of only "+nn+" nodes.");
        // all edges are sorted, and all clique-subsets are generated
        // in sorted order. So we don't have to search much. The
        // subsets generated are spaced out, the edges are by
        // definition dense, so for each new subset generated, we
        // iterate until we find the edge or until the list is over. A
        // possible speedup would be to implement an order for edges
        // which would allow us to stop earlier.
        Iterator iter = edges.iterator();
        KEdge tmpEdge2;

        // basically, what we have to check is whether all
        // "arity"-size subsets of "indices" are exactly one edge in
        // this hypergraph.

        
        // form all arity-subsets of nodes (but only those
        // whose node numbers are in lexicographic order)
        int indices[] = new int[arity]; 
        int data[] = new int[arity];
        // for the indices of the subset elements
        
        // initialize indices
        for (int i = 0; i<arity; i++)
            indices[i]=i;
        
        int i_ptr = arity-1;
        // now form subsets 
        
        do {
            if (DEBUG) { 
                System.out.print("(cliquefinding) Subset: ");
                for (int m=0; m<arity; m++)
                    System.out.print(subset[indices[m]]+" ");
                System.out.println();
            }

            for (int i=0; i<arity; i++) 
                data[i]=subset[indices[i]];
            
            
            // cycle through edges to see if one is missing
            do {
                if (!iter.hasNext()) return false; // is no clique
                tmpEdge2=(KEdge)iter.next();
                if (DEBUG) System.out.println("comparing to: "+tmpEdge2.toString());
            }
            while (!tmpEdge2.equals(data));

            // next subset
            i_ptr=arity-1;
            indices[i_ptr]++;
            
            // if out of bounds, push up the carry-overs
            for (i_ptr=arity-1;i_ptr>0;i_ptr--) {
                if (indices[i_ptr]==nn-(arity-1-i_ptr)) {// index at position k cannot be larger than k+(nn-arity)
                    if (i_ptr==0) break;
                    else {
                        indices[i_ptr-1]++;
                        indices[i_ptr]=-1;
                    }
                }
            }
            for(i_ptr=1; i_ptr<arity; i_ptr++) {
                if (indices[i_ptr]==-1)
                    indices[i_ptr]=indices[i_ptr-1]+1;
            }
        }
        while (indices[0]<nn-(arity-1));
        
        
        return true;

    }

    /** returns true if there is an edge that has both nodes as element */
    public boolean connected(int node1, int node2) {
	KEdge tmpEdge;
	if (node1==node2) return true;
	for (Iterator iter=edges.iterator(); iter.hasNext();) {
	    tmpEdge=(KEdge)iter.next();
	    if (tmpEdge.contains(node1) && tmpEdge.contains(node2))
		return true;
	}
	return false;
    }

    /** returns true if there is an edge that has node1 but NOT
        node2. This is NOT the inverse of "connected"!! */
    public boolean disconnected(int node1, int node2) {
	KEdge tmpEdge;
	for (Iterator iter=edges.iterator(); iter.hasNext();) {
	    tmpEdge=(KEdge)iter.next();
	    if (tmpEdge.contains(node1) && !tmpEdge.contains(node2)) {
		System.out.println("Disconnection check:  Edge "+tmpEdge+" disconnects nodes "+node1+" and "+node2);
		return true;
	    }
	}
	return false;
    }



    public String toString() {
        String result="Arity: "+arity+"\n";
        for (int i =0; i<edges.size(); i++)
            result += (new Integer(i).toString())+": "+ edges.elementAt(i).toString()+"\n";
        return result;
    }


    /** binomial coefficient of n and k */
    static private int binomial(int n, int k) {
        if (n<k) return -1; // error

        int result =1;
        for(int i = 0; i< k; i++)
            result *= n-i;
        result /= fact(k);
        return result;
    }

    /** factorial function */
    public static int fact(int x) {
	int result=1;
	while(x>1) result*=x--;
	return result;
    }
}


/** a helper class to store edges in Vectors. The constructor makes a
    copy of the passed integer array, i.e., it can be overwritten in
    the calling function. */
class KEdge {
    private int arity;
    private int edge[];

    public KEdge(int[] edgeP) {
        arity = edgeP.length;
        edge = new int[arity];
        System.arraycopy(edgeP,0,edge,0,arity);
        insertionSort();
    }

    public int[] getEdge() {
        return edge;
    }

    public int getArity() {
	return arity;
    }

    /** returns node index at dimension dim */
    public int nodeAt(int dim) {
        if (dim>=arity) 
	    throw new IndexOutOfBoundsException(new Integer(dim).toString()+ ">="+ new Integer(arity).toString());
        return edge[dim];
    }


    /** compares this edge with an edge given implicitly by its node
        IDs. IDs must be lexicographically sorted */
    public boolean equals(int[] edge) {
        int thisedge[]=getEdge();
        if (thisedge.length!=edge.length) return false;
        for(int i=0; i<edge.length; i++)
            if (thisedge[i]!=edge[i]) return false;
        return true;
    }

    /** for the contains-function */
    public boolean equals(Object theEdge) {
	return equals((KEdge) theEdge);
    }

    public boolean equals(KEdge edge2p) {
        int edge2[] = edge2p.getEdge();
        if (edge2.length!=arity) return false;
        for(int i=0; i<arity; i++)
            if(edge[i]!=edge2[i]) return false;
        return true;
    }
    
    /** returns true if this edge is lexicographically smaller than
        the parameter */
    public boolean lessThanOrEqual(KEdge edge2p) {
        int edge2[] = edge2p.getEdge();
        if (edge2.length!=arity) throw new RuntimeException("Comparing two edges of different size!");
        for(int i=0; i<arity; i++) {
            if(edge[i]<edge2[i]) return true;
            if(edge[i]>edge2[i]) return false;
        }
        return true;
        
    }
    
    /** checks if a certain node (index) is contained in this
	edge. Right now it's a linear search, but could be made binary
	with little difficulty */
    public boolean contains(int node) { 
	for (int i=0; i<arity; i++) 
	    if(node==edge[i]) return true;
	return false;
    }

    /** returns true if this edge is a subset of the edge in the
        parameter. It is as usual assumed that the nodes in the edges are both sorted. */
    public boolean isSubsetOf(KEdge superEdge) {

	int superArity=superEdge.getArity();
	int superEdgeArray[] = superEdge.getEdge();
	if (superArity<arity) return false;
	int j=0,i;
	for (i =0; i<arity && j<superArity; i++) {
	    while(j<superArity && edge[i]!=superEdgeArray[j]) 
		j++;
	}
	if(j==superArity) return false;
	else return true;
    }
    
    /** performs an insertion sort of the node numbers in this array,
        in order to keep only sorted edges */
    private void insertionSort() {
        int tmpNode,j;
        for (int i = 1 ; i<arity; i++) {
            j = i-1;
            tmpNode = edge[i];
            while (j>=0 && tmpNode<edge[j]) {
                edge[j+1]=edge[j];
                j--;
            }
            edge[j+1]=tmpNode;
        }
    }

    public String toString() {
        String result = "(";
        for (int i =0; i<arity-1; i++)
            result+= edge[i]+" ";
        result += edge[arity-1]+")";
        return result;
    }
}
